//***************************************************************************//
//* FRIDA: fast reliable interactive data analysis                          *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                 *//
//* http://apps.jcns.fz-juelich.de                                          *//
//***************************************************************************//

//! \file  manip.cpp
//! \brief NManip: manipulate data (select, average, rearrange, ...)

#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <iostream>
#include <string>
#include <vector>
#include <algorithm>
#include <boost/format.hpp>
#include <gsl/gsl_sort.h>

#include <trivia/vector_ops.hpp>
#include <read-plus/ask.hpp>

#include "defs.h"
#include "list.h"
#include "olf.h"
#include "mem.h"
#include "jsel.h"
#include "var.h"
#include "zentry.h"
#include "expr.h" // for sort, symmetrize
#include "manip.h"
#include "xax_lex.h"

using boost::format;

// local routines:
namespace NManip {
    void ScaRemoveConstantZ( POlo f );
}


//***************************************************************************//
//* restrict forthcoming operations                                         *//
//***************************************************************************//

//! Restrict forthcoming operations

void NManip::ProtectSpecs( bool on )
{
    NOlm::IterateO fiter;
    string sel;
    if( on )
        sel = wask( "Protect which spectra from forthcoming operations" );
    else
        sel = "-";
    while( POlo fio = fiter() ) {
        CList lis( sel, 0, fio->nJ()-1 );
        for( uint j=0; j<fio->nJ(); ++j )
            fio->V[j]->protect = lis.contains(j);
        fio->lDoc.push_back( "m/ " + sel );
    }
}


//***************************************************************************//
//* manipulations on data points                                            *//
//***************************************************************************//

//! Delete (or retain) points according to list of indices.

void NManip::PtsSelect( bool sel_del )
{
    NOlm::IterateD fiter;
    static string pSel = "";
    pSel = wask( string( sel_del ? "Delete" : "Retain" ) +
                 " which points", pSel );
    CList pLis( pSel );

    while( POld fin = fiter() ) {
        POld fout( fin->new_old() );
        fout->lDoc.push_back( "mp" + string( sel_del ? "d" : "r" )
                              + " " + pSel );

        for ( uint j=0; j<fin->nJ(); j++ ) {
            PSpec sin = fin->VS(j);
            PSpec sout( new CSpec );
            sout->copy_z_base( sin );
            pLis.evaluate( 0, sin->size()-1 );
            for ( uint i=0; i<sin->size(); i++ ) {
                if ( sel_del ^ pLis.contains(i) ) {
                    if( sin->dy.size() )
                        sout->push_xyd( sin->x[i], sin->y[i], sin->dy[i] );
                    else
                        sout->push_xy( sin->x[i], sin->y[i] );
                }
            }
            if ( sout->size() )
                fout->V.push_back( sout );
        }
        if ( !fout->V.size() )
            throw "no data point left";
        NOlm::mem_store( fout, fiter.k() );
    }
}


//! Bin points.

void NManip::PtsAvge()
{
    NOlm::IterateD fiter;

    static string sel;
    sel = wask( "Start bins at points", sel );
    if ( sel=="" || sel=="-" )
        return;
    
    while( POld fin = fiter() ) {
        POld fout( fin->new_old() );
        fout->lDoc.push_back("mpa " + sel);

        for( uint j=0; j<fin->nJ(); j++ ) {
            PSpec sin = fin->VS(j);
            PSpec sout( new CSpec() );
            sout->copy_z_base( sin );

            CList groups( sel, 0, sin->size()-1 );
            if ( !groups.size() )
                throw "BUG: unexpected empty bin list";
            if ( groups.V[0]!=0 )
                throw "First bin must start at 0";
            vector<uint> breaks( groups.V );
            breaks.push_back( sin->size() );

            for ( uint iv=0; iv<breaks.size()-1; ++iv ) {
                uint igi = breaks[iv];
                uint igf = breaks[iv+1];
                double xg = 0, yg = 0, vg = 0;
                for ( uint i=igi; i<igf; ++i ) {
                    xg += sin->x[i];
                    yg += sin->y[i];
                    if( sin->dy.size() )
                        vg += SQR( sin->dy[i] );
                }
                uint ng = igf - igi;
                if( sin->dy.size() )
                    sout->push_xyd(xg/ng, yg/ng, sqrt(vg)/ng );
                else
                    sout->push_xy(xg/ng, yg/ng );
            }
            fout->V.push_back(sout);
        }
        NOlm::mem_store( fout, fiter.k() );
    }
}


//! Bin points.

void NManip::PtsAvgeFixed()
{
    NOlm::IterateD fiter;
    static int ng=2;
    ng = iask( "Number of points per bin", ng );
    if ( ng<1 )
        throw "invalid choice";
    
    while( POld fin = fiter() ) {
        POld fout( fin->new_old() );
        fout->lDoc.push_back( "mpaf " + S(ng) );

        for( uint j=0; j<fin->nJ(); j++ ) {
            PSpec sin = fin->VS(j);
            PSpec sout( new CSpec() );
            sout->copy_z_base( sin );

            if( sin->size()<ng )
                throw "not enough points in spectrum " + S(j);

            for ( uint iv=0; iv<sin->size()/ng; ++iv ) {
                double xg = 0, yg = 0, vg = 0;
                for ( uint i=iv*ng; i<(iv+1)*ng; ++i ) {
                    xg += sin->x[i];
                    yg += sin->y[i];
                    if( sin->dy.size() )
                        vg += SQR( sin->dy[i] );
                }
                if( sin->dy.size() )
                    sout->push_xyd(xg/ng, yg/ng, sqrt(vg)/ng );
                else
                    sout->push_xy(xg/ng, yg/ng );
            }
            fout->V.push_back(sout);
        }
        NOlm::mem_store( fout, fiter.k() );
    }
}


//! Bin points to keep errors below a certain bound.

void NManip::PtsAvgeByErr( const string& subcmd )
{
    NOlm::IterateD fiter;
    bool mode_abs;
    if      ( subcmd=="a" )
        mode_abs = true;
    else if ( subcmd=="r" )
        mode_abs = false;
    else
        throw "invalid subcommand";

    static double errbound;
    errbound = dask( "keep error below", errbound );
    
    while( POld fin = fiter() ) {
        POld fout( fin->new_old() );
        string groupinfo;
        for ( uint j=0; j<fin->nJ(); j++ ) {
            PSpec sin = fin->VS(j);
            PSpec sout( new CSpec() );
            sout->copy_z_base( sin );
            if( !sin->dy.size() )
                throw "no errors available";
            double xg = 0, yg = 0, vg = 0;
            uint ng = 0, istart=0;
            for( uint i=0; i<sin->size(); ++i ) {
                if ( !mode_abs && sin->y[i]<0 )
                    throw "negative counts not compatible with"
                                  " relative error bounds";
                xg += sin->x[i];
                yg += sin->y[i];
                vg += SQR( sin->dy[i] );
                ++ng;
                double dg = sqrt(vg);
                if( (  mode_abs && dg/ng<errbound ) ||
                    ( !mode_abs && yg>0 && dg/yg<errbound ) ) {
                    sout->push_xyd( xg/ng, yg/ng, dg/ng );
                    groupinfo += S(istart) + ",";
                    xg = 0; yg = 0; vg = 0; ng = 0;
                    istart = i+1;
                }
            }
            if( sout->size() )
                fout->V.push_back(sout);
        }
        string docline = "mpae" + subcmd + " " + S(errbound);
        if( fin->nJ()==1 )
            docline += " #=> groups " + groupinfo.substr(0,groupinfo.size()-1);
        fout->lDoc.push_back( docline );
        if( fout->nJ() )
            NOlm::mem_store( fout, fiter.k() );

    }
}


//! Sort spectrum according to given expression.

void NManip::PtsSort()
{
    NOlm::IterateD fiter;
    string expr = sask("Sort points according to");
    if (expr=="") return;
    PNode T = user_xaxparse( expr.c_str() );
    if( T->has_dummy() )
        throw "dummy argument t not allowed when operating on data";

    while( POld fin = fiter() ) {
        uint k, n, i;
        k = fiter.k();
        POld fout( fin->new_old() );
        fout->lDoc.push_back("mpo "+expr);

        for ( uint j=0; j<fin->nJ(); j++ ) {
            PSpec sin = fin->VS(j);
            PSpec sout( new CSpec );
            sout->copy_z_base( sin );

            n = sin->size();
            vector<double> v(n);
            T->tree_vec_val( &v, 0, k, j );
            size_t P[n];
            gsl_sort_index (P, &(v[0]), 1, n);
            sout->resize( n, sin->dy.size() );
            for (i=0; i<n; ++i) {
                sout->x[i] = sin->x[P[i]];
                sout->y[i]  = sin->y[P[i]];
                if( sin->dy.size() )
                    sout->dy[i] = sin->dy[P[i]];
            }
            fout->V.push_back( sout );
        }
        NOlm::mem_store( fout, fiter.k() );
    }
}


//! Average points that have the same x.

void NManip::PtsAvgeEq(void)
{
    NOlm::IterateD fiter;
    COld fout;
    while( POld fin = fiter() ) {
        POld fout( fin->new_old() );
        fout->lDoc.push_back("mpq # average points when x equal");

        for ( uint j=0; j<fin->nJ(); j++ ) {
            const PSpec sin = fin->VS(j);
            PSpec sout( new CSpec );
            sout->copy_z_base( sin );

            uint n = sin->size();
            uint ii=0;
            for ( uint i=0; i<n; i=ii ) {
                double xg = sin->x[i];
                double yg = sin->y[i];
                double eg;
                if( sin->dy.size() )
                    eg = SQR(sin->dy[i]);
                uint ng = 1;
                for ( ii=i+1; ii<n; ++ii ) {
                    if (sin->x[ii] != xg) break;
                    yg += sin->y[ii];
                    if( sin->dy.size() )
                        eg += SQR( sin->y[ii] );
                    ++ng;
                }
                if( sin->dy.size() )
                    sout->push_xyd(xg, yg/ng, sqrt(eg)/ng );
                else
                    sout->push_xy(xg, yg/ng);
            }
            fout->V.push_back( sout );
        }
        NOlm::mem_store( fout, fiter.k() );
    }
}


//! Symmetrize spectrum.

void NManip::PtsSymmetrize()
{
    NOlm::IterateD fiter;
    static string expr = "0";
    expr = sask( "Assume symmetry with mirror positioned at", expr );
    if ( expr=="" ) return;
    PNode T = user_xaxparse( expr.c_str() );

    while( POld fin = fiter() ) {
        uint k = fiter.k();
        POld fout( fin->new_old() );
        fout->lDoc.push_back("mpsym "+expr);

        uint n, i, il, ih, nl, nh;
        for (uint j=0; j<fin->nJ(); j++) {
            PSpec sin = fin->VS(j);
            PSpec sout( new CSpec );
            sout->copy_z_base( sin );
            if( !triv::is_ascending( sin->x ) )
                throw "not ascending";
            double step;
            if( !triv::is_equidist( &step, sin->x ) )
                throw "not equidistant";
            double centre;
            T->tree_point_val( &centre, 0, k, j );
            il = (int) ( (centre - sin->x[0])/step );
            ih = il + 1;
            n = sin->size();
            nl = il + 1;
            nh = n - ih;
            if( nl<1 || nh<1 )
                throw "nothing to symmetrize";
            for (i=0; i<(nl<nh?nl:nh); ++i) {
                if( sin->dy.size() )
                    sout->push_xyd( ( -sin->x[il-i] + sin->x[ih+i] ) / 2,
                                    ( sin->y[il-i] + sin->y[ih+i] ) / 2, 
                                    sqrt( SQR(sin->dy[il-i]) +
                                          SQR(sin->dy[ih+i]) ) / 2 );
                else
                    sout->push_xy( ( -sin->x[il-i] + sin->x[ih+i] ) / 2,
                                   ( sin->y[il-i] + sin->y[ih+i] ) / 2 );
            }
            if( nl<nh ){
                for (i=nl; i<nh; ++i)
                    if( sin->dy.size() )
                        sout->push_xyd( sin->x[ih+i], sin->y[ih+i],
                                        sin->dy[ih+i] );
                    else
                        sout->push_xy( sin->x[ih+i], sin->y[ih+i] );
            } else {
                for (i=nh; i<nl; ++i)
                    if( sin->dy.size() )
                        sout->push_xyd( -sin->x[il-i], sin->y[il-i],
                                        sin->dy[il-i] );
                    else
                        sout->push_xy( -sin->x[il-i], sin->y[il-i] );
            }
            sout->z = sin->z;
            fout->V.push_back(sout);
        }
        NOlm::mem_store( fout, fiter.k() );
    }
}


//***************************************************************************//
//* manipulations on spectra                                                *//
//***************************************************************************//

//! Remove z variable that does not vary with j.

void NManip::ScaRemoveConstantZ( POlo fio )
{
    for( int iz=fio->nZ()-1; iz>=0; --iz ){
        uint nj = fio->nJ();
        if( nj<1 )
            throw "file contains no spectrum";
        double z = fio->z(0,iz);
        bool z_is_const = true;
        for( uint j=1; j<nj; ++j ){
            if( fio->z(j,iz) != z ){
                z_is_const = false;
            }
        }
        if( z_is_const ){
            printf( " constant z%i becomes r%zu\n", iz, fio->RPar.size() );
            fio->RPar.push_back( CParam( fio->ZCo[iz], z ) );
            fio->remove_z( iz );
        }
    }
}


//! Delete (or retain) spectra according to list of indices.

void NManip::ScaSelect( bool sel_del )
{
    CList JSel, JSelSorted;
    static string jSel = "";
    JSelAsk( string( sel_del ? "Delete" : "Retain" ) +
                   " which spectra", &jSel, &JSel );

    NOlm::IterateO fiter;
    while( POlo fin = fiter() ) {
        POlo fout( fin->new_olo() );
        JSel.evaluate( 0, fin->nJ()-1 );
        fout->lDoc.push_back( "ms" + string( sel_del ? "d" : "r" )
                              + " " + JSel.str());
        for( int j=0; j<fin->nJ(); ++j ) {
            if( sel_del ^ JSel.contains( j ) ){
                fout->V.push_back( fin->new_zentry( j ) );
            }
        }
        ScaRemoveConstantZ( fout );
        NOlm::mem_store( fout, fiter.k() );
    }
}


//! Bin spectra.

void NManip::ScaAvge()
{
    NOlm::IterateO fiter;
    CList JSel;
    static string jSel = "";
    JSelAsk( "Start groups at spectra", &jSel, &JSel );
    while( POlo fin = fiter() ) {
        POld fd = P2D( fin ); 
        POlc fc = P2C( fin ); 
        POlo fout( fin->new_olo() );
        JSel.evaluate( 0, fin->nJ()-1 );
        if ( JSel.V.size()<1 || JSel.V[0]!=0 )
            throw "spectrum selection must contain 0";
        fout->lDoc.push_back("msa " + JSel.str());
        for( uint iv=0; iv<JSel.size(); iv++ ) {
            uint ji = JSel.V[iv];
            uint jf = ( iv<JSel.size()-1 ? JSel.V[iv+1] : fin->nJ() );
            uint nz = fin->nZ();
            vector<double> zout( nz );
            for( uint iz=0; iz<nz; ++iz ){
                uint mz = 0;
                double z = 0;
                for ( uint jj=ji; jj<jf; jj++ ) {
                    z += fin->z(jj,iz);
                    ++mz;
                }
                zout[iz] = z/mz;
            }
            if( fd ) {
                PSpec sout( new CSpec() );
                bool with_dy = fd->VS(0)->dy.size();
                uint n = fd->nPts(ji);
                sout->resize( n, with_dy );
                // x grids must be equal:
                sout->x = fd->VS(ji)->x;
                for ( uint jj=ji+1; jj<jf; jj++ ){
                    if ( fd->nPts(jj) != fd->nPts(ji) )
                        throw "spectrum " + S(jj) +
                            " has other size than spectrum " + S(ji);
                    for( uint i=0; i<fd->nPts(jj); ++i ){
                        if ( fd->VS(jj)->x[i] != sout->x[i] )
                            throw "spectrum " + S(jj) + " has other x[" +
                                S(i) + "] than spectrum 0";
                    }
                }
                // bin y:
                for ( uint i=0; i<n; i++ ){
                    uint mj = 0;
                    double ym = 0;
                    double vm = 0;
                    for ( uint jj=ji; jj<jf; jj++) {
                        ym += fd->VS(jj)->y[i];
                        if( with_dy )
                            vm += SQR(fd->VS(jj)->dy[i] );
                        ++mj;
                    }
                    sout->y[i] = mj==0 ? 0 : ym/mj;
                    if( with_dy )
                        sout->dy[i] = mj==0 ? 0 : sqrt(vm)/mj;
                }
                sout->z = zout;
                fout->V.push_back( sout );
            } else if ( fc ) {
                PCurve cout( new CCurve );
                uint np = fc->nP;
                vector<double> p(np);
                for ( uint ip=0; ip<np; ip++ ){
                    uint mj = 0;
                    double pm = 0;
                    for ( uint jj=ji; jj<jf; jj++) {
                        pm += fc->VC(jj)->P[ip];
                        ++mj;
                    }
                    p[ip] = mj==0 ? 0 : pm/mj;
                }
                cout->P = p;
                cout->z = zout;
                cout->fixed.resize( np, false );
                fout->V.push_back( cout );
            }
        }
        ScaRemoveConstantZ( fout );
        NOlm::mem_store( fout, fiter.k() );
    }
}


//! Merge spectra, concatenating points.

void NManip::ScaJoin()
{
    NOlm::IterateD fiter;
    CList JSel;
    static string jSel = "";
    JSelAsk( "Start groups at spectra", &jSel, &JSel );

    while( POld fin = fiter() ) {
        POld fout( fin->new_old() );
        JSel.evaluate( 0, fin->nJ()-1 );
        if (JSel.V.size()<1 || JSel.V[0]!=0)
            throw "spectrum selection must contain 0";
        fout->lDoc.push_back("msj " + JSel.str());
        cerr << "WARNING: z just taken from first spectrum of each group\n";
        for ( uint iv=0; iv<JSel.size(); iv++ ) {
            uint ji, jf;
            ji = JSel.V[iv];
            jf = ( iv<JSel.size()-1 ? JSel.V[iv+1] : fin->nJ() );
            PSpec sout( new CSpec( *(fin->VS(ji)) ) );
            bool with_dy = fin->VS(ji)->dy.size();
            for ( uint jj=ji+1; jj<jf; jj++){
                for( uint i=0; i<fin->VS(jj)->size(); ++i )
                    if ( with_dy )
                        sout->push_xyd( fin->VS(jj)->x[i], fin->VS(jj)->y[i],
                                        fin->VS(jj)->dy[i]);
                    else
                        sout->push_xy( fin->VS(jj)->x[i], fin->VS(jj)->y[i] );
            }
            fout->V.push_back(sout);
        }
        ScaRemoveConstantZ( fout );
        NOlm::mem_store( fout, fiter.k() );
    }
}


//! After each span, insert given number of copies.

void NManip::ScaSpawn()
{
    NOlm::IterateO fiter;
    static int njj = 2;
    int njjIn = iask("output will have how many copies of input spectrum", njj);
    if (njjIn<1) return;
    njj = njjIn;

    while( POlo fin = fiter() ) {
        POlo fout( fin->new_olo() );
        fout->lDoc.push_back( "ms* " + S(njj) );

        fout->ZCo.insert( fout->ZCo.begin(), CCoord("no-in-spawn", ""));

        for( uint jj=0; jj<njj; ++jj ){
            for( uint j=0; j<fin->nJ(); j++ ){
                PZentry eout = fin->new_zentry( j );
                eout->z.insert( eout->z.begin(), jj );
                fout->V.push_back( eout );
            }
        }
        NOlm::mem_store( fout, fiter.k() );
    }
}


//! Exchange x and z.

void NManip::ScaExch()
{
    NOlm::IterateD fiter;
    while( POld fin = fiter() ) {
        POld fout( fin->new_old() );

        uint izco = fin->ZCo.size() - 1;
        if ( izco==(uint)-1 )
            throw "no input z coordinate";
        fout->lDoc.push_back(
            "msx # exchanging x=" + fin->xco.str_compact() +
            " with z" + S(izco) + "=" + fin->ZCo[izco].str_compact() );
        fout->xco = fin->ZCo[izco];
        fout->ZCo[izco] = fin->xco;
		
        if ( fin->nJ()<=0 )
            throw "no spectra in file";

        vector<double> zcommon, xcommon;
        uint ji, jf;

        for ( ji=0; ji<fin->nJ(); ) {
            zcommon.clear();
            for ( uint i=0; i<izco; ++i)
                zcommon.push_back(fin->V[ji]->z[i]);

            for ( jf=ji+1; jf<fin->nJ(); jf++ )
                for ( uint i=0; i<izco; ++i)
                    if (fin->V[jf]->z[i] != zcommon[i])
                        goto end_group;
        end_group:

            // build common grid:
            xcommon.clear();
            for ( uint j=ji; j<jf; ++j )
                for ( uint i=0; i<fin->VS(j)->size(); ++i )
                    xcommon.push_back( fin->VS(j)->x[i] );
            sort( xcommon.begin(), xcommon.end() );
            triv::unique(&xcommon, 1e-180, 1e-20);

            printf("group of spectra %u .. %u has %zu different x\n", 
                   ji, jf-1, xcommon.size());

            uint ii = 0; // a guess
            for ( uint i=0; i<xcommon.size(); ++i ) {
                PSpec sout( new CSpec );
                
                sout->z = zcommon;
                sout->z.push_back(xcommon[i]);

                for ( uint j=ji; j<jf; ++j ) {
                    PSpec sin = fin->VS( j );
                    bool with_dy = sin->dy.size();
                    if( sin->x[ii]!=xcommon[i] ){ // guess failed
                        for (ii=0; ii<sin->size(); ++ii)
                            if( sin->x[ii]==xcommon[i] )
                                break;
                    }
                    if( ii<sin->size() ){
                        if( with_dy )
                            sout->push_xyd( sin->z[izco], sin->y[ii],
                                           sin->dy[ii] );
                        else
                            sout->push_xy( sin->z[izco], sin->y[ii] );
                    }
                }
                fout->V.push_back( sout );
                ++ii;
            }
            ji = jf;
        }
        NOlm::mem_store( fout, fiter.k() );
    }
}


//! Sort spectra according to expression.

void NManip::ScaSortByExpr()
{
    NOlm::IterateO fiter;
    string expr = sask("Sort spectra according to");
    if (expr=="") return;
    PNode T = user_xaxparse( expr.c_str() );

    while( POlo fin = fiter() ) {
        POlo fout( fin->new_olo() );
        fout->lDoc.push_back( "mso "+expr );

        uint nj = fin->nJ();
        vector<double> v(nj);
        for (uint j=0; j<nj; j++)
            T->tree_point_val( &(v[j]), 0, fiter.k(), j );
        
        size_t pos[nj];
        gsl_sort_index( pos, &(v[0]), 1, nj );
                
        for (uint j=0; j<nj; j++)
            fout->V.push_back( fin->new_zentry( pos[j] ) );

        NOlm::mem_store( fout, fiter.k() );
    }
}


//! Sort spectra according to full z vector.

void NManip::ScaSortByZ()
{
    NOlm::IterateO fiter;
    while( POlo fin = fiter() ) {
        POlo fout( fin->new_olo() );
        for (uint j=0; j<fin->nJ(); j++)
            fout->V.push_back( fin->new_zentry( j ) );
        sort( fout->V.begin(), fout->V.end(), CompareZ );
        NOlm::mem_store( fout, fiter.k() );
    }
}


//! Change order of z coordinates.

void NManip::ZExchange()
{
    NOlm::IterateO fiter;
    int nzmin, nzmax = 0;
    char mod;
    string com;
    int num1, num2;

    while( POlo fin = fiter() ) {
        nzmax = max( (int)nzmax, (int)fin->ZCo.size() );
    }
    if       ( nzmax<=1 )
        throw "nothing to exchange";
    else if( nzmax==2 ){
        mod = 'r';
        printf( "exchanging two z coordinates\n" );
        com = "exchanged two z coordinates";
        num1 = 1;
        nzmin = 2;
    } else {
        string sel;
        sel = wask( "rotate(r,rr,...), revert(m), exchange(<number>)" );
        if      ( sel[0]=='r' ){
            mod = 'r';
            num1 = 1;
            while( sel.length() > num1 ){
                if( sel[num1++]!='r' )
                    throw "invalid selection";
            }
            com = "rotated z coordinates by " + S(num1);
            nzmin = 2;
        } else if( sel=="m" ){
            mod = 'm';
            com = "reverted z coordinates";
            nzmin = 2;
        } else {
            if( sscanf(sel.c_str(), "%d", &num1)!=1 )
                throw "invalid selection";
            if( num1>=nzmax )
                throw "no more than " + S(nzmax) + " z coordinates";
            num2 = iask( "exchange with coordinate" );
            if( num2>=nzmax )
                throw "no more than " + S(nzmax) + " z coordinates";
            else if( num1==num1 )
                throw "nothing to exchange";
            mod = 'x';
            com = "exchanged z coordinates "+S(num1)+" and "+S(num2);
            nzmin = max( num1, num2 )+1;
        }
    }
                
    fiter.reset();
    while( POlo fin = fiter() ) {
        POlo fout( fin->new_olo() );
        fout->lDoc.push_back( com );
        for (uint j=0; j<fin->nJ(); j++)
            fout->V.push_back( fin->new_zentry( j ) );

        int nz = fout->ZCo.size();
        if      ( nz<nzmin ){
            cerr << "WARNING nz<nzmin\n";
        } else if( mod=='r' ){
            for( int irot=0; irot<num1; ++irot ){
                for( int iz=1; iz<nz; ++iz ){
                    fout->ZCo[iz] = fin->ZCo[iz-1];
                }
                fout->ZCo[0] = fin->ZCo[nz-1];
            }
        } else if( mod=='m' ){
            for( int iz=0; iz<nz; ++iz ){
                fout->ZCo[iz] = fin->ZCo[nz-1-iz];
            }
        } else { // mod=='x'
            fout->ZCo[num1] = fin->ZCo[num2];
            fout->ZCo[num2] = fin->ZCo[num1];
        }

        for ( uint j=0; j<fin->V.size(); j++ ) {
            if      ( nz<nzmin ){
                // do nothing
            } else if( mod=='r' ){
                for( int irot=0; irot<num1; ++irot ){
                    for( int iz=1; iz<nz; ++iz )
                        fout->V[j]->z[iz] = fin->V[j]->z[iz-1];
                    fout->V[j]->z[0] = fin->V[j]->z[nz-1];
                }
            } else if( mod=='m' ){
                for( int iz=0; iz<nz; ++iz ){
                    fout->V[j]->z[iz] = fin->V[j]->z[nz-1-iz];
                }
            } else  { // mod=='x'
                fout->V[j]->z[num1] = fin->V[j]->z[num2];
                fout->V[j]->z[num2] = fin->V[j]->z[num1];
            }
        }
        NOlm::mem_store( fout, fiter.k() );
    }
}


//! Delete z coordinate.

void NManip::ZDelete()
{
    NOlm::IterateO fiter;
    string sel;
    sel = wask( "List of z coordinates to be removed" );
    if ( sel=="-" )
        return;
	
    while( POlo fin = fiter() ) {
        POlo fout( fin->new_olo() );
        fout->lDoc.push_back( "mz- "+sel );

        if ( !fin->nJ() )
            throw "BUG: no spectra";
        for ( uint j=0; j<fin->nJ(); j++ )
            fout->V.push_back( fin->new_zentry( j ) );

        CList lis( sel, 0, fin->ZCo.size()-1 );
        lis.sortReverse();

        uint nRPar = fout->RPar.size();
        for ( uint iv=0; iv<lis.size(); ++iv ) {
            uint iz = lis.V[iv];
            fout->RPar.insert( fout->RPar.begin()+nRPar, fout->ZCo[iz] );
            fout->ZCo.erase( fout->ZCo.begin()+iz );
            double zval = 0;
            for ( uint j=0; j<fin->nJ(); j++ ) {
                zval += fout->V[j]->z[iz];
                fout->V[j]->z.erase( fout->V[j]->z.begin()+iz );
            }
            fout->RPar[nRPar].val = zval / fin->nJ();
        }
        NOlm::mem_store( fout, fiter.k() );
    }
}


//! Break file into several files, according to z0.

void NManip::ScaBreak()
{
    NOlm::IterateO fiter;
    while( POlo fin = fiter() ) {
        if ( !(fin->ZCo.size()) )
            throw "no z coordinate";
        CCoord zco = fin->ZCo[0];

        // intermediate file, on which output files will be based:
        POlo ftmp( fin->new_olo() );
        ftmp->V.clear();
        ftmp->lDoc.push_back( "msb # eliminating " + zco.str_compact() );
        ftmp->ZCo.erase( ftmp->ZCo.begin() );
        uint jout = 0;

        for ( uint j=0; j<fin->nJ(); ) {
            POlo fout( ftmp->new_olo() );
            double zval = ftmp->z(j,0);
            fout->RPar.push_back( CParam( zco, zval ) );
            do {
                PZentry eout( fin->new_zentry(j) );
                eout->z.erase( eout->z.begin() );
                fout->V.push_back( eout );
            } while ( ++j<fin->nJ() && fin->z(j,0)==zval );
            fout->name += "_" + S(jout++);
            NOlm::mem_store( fout );
        }
        if( NOlm::overwrite )
            NOlm::mem_del( fiter.k() );
    }
}


//***************************************************************************//
//* manipulations on files                                                  *//
//***************************************************************************//

//! Merge files, concatenating spectra/curves.

void NManip::FilMerge( const string& opts )
{
    NOlm::IterateO fiter;
    // parse options:
    bool add_zk = false;
    if ( opts=="+" )
        add_zk = true;

    // output file type according to first input file:
    POlo fin = fiter();
    POlo fout;
    if        ( POld fd = P2D(fin) ) {
        fout = POld( new COld );
        while ( fin = fiter() ){
            if ( !P2D(fin) )
                throw "cannot merge data and curve file";
        }
    } else if ( POlc fc = P2C(fin) ) {
        fout = POlc( new COlc );
        P2C(fout)->expr = fc->expr;
        while ( fin = fiter() ){
            if ( !P2C(fin) )
                throw "cannot merge curve and data file";
            if ( P2C(fin)->expr != fc->expr )
                throw "different curve definitions";
        }
    } else 
        throw "BUG: unexpected POlo type";

    // check and set coordinates
    fiter.reset();
    fin = fiter();
    fout->xco = fin->xco;
    fout->yco = fin->yco;
    fout->ZCo = fiter()->ZCo;
    if ( POlc fc = P2C(fin) )
        P2C(fout)->PCo = fc->PCo;
    while ( fin = fiter() ){
        if ( fin->xco != fout->xco )
            throw "different x coordinates";
        if ( fin->yco != fout->yco )
            throw "different y coordinates";
        if ( fin->ZCo.size() != fout->ZCo.size() )
            throw "different numer of z coordinates";
        for ( uint i=0; i<fout->ZCo.size(); ++i )
            if ( fin->ZCo[i] != fout->ZCo[i] )
                throw "different z"+S(i)+" coordinate";
        if ( POlc fc = P2C(fin) )
            if ( fc->PCo.size() != P2C(fout)->PCo.size() )
                throw "different numer of curve parameters";
    }

    // check RPar, and prepare move to z if they differ:
    fiter.reset();
    vector<CParam> RPar( fiter()->RPar );
    uint nR = RPar.size();
    vector<bool> RParDiff( nR, false );    
    while ( (fin = fiter()) ) {
        if ( fin->RPar.size() != nR )
            throw "different numer of r parameters";
        for ( uint i=0; i<nR; ++i )
            if (fin->RPar[i].Co != RPar[i].Co )
                throw "different r parameter";
        for ( uint i=0; i<nR; ++i )
            if (fin->RPar[i].val != RPar[i].val )
                RParDiff[i] = true;
    }
    auto zinsert = fout->ZCo.begin();
    for ( uint i=0; i<nR; ++i ) {
        if( RParDiff[i] )
            fout->ZCo.insert( zinsert++, RPar[i].Co );
        else
            fout->RPar.push_back( RPar[i] );
    }
    if ( add_zk || !fout->ZCo.size() ) {
        // there must be at least one z coordinate
        fout->ZCo.insert( fout->ZCo.begin(),
                          CCoord("number-of-input-file", "") );
        add_zk = true;
    }

    // file name:
    string fnam;
    fiter.reset();
    fnam = fiter()->name;
    while ( fin = fiter() ) {
        uint n2=fin->name.size();
        if ( n2<fnam.size() )
            fnam.erase( fnam.begin()+n2, fnam.end() );
        for ( uint i=0; i<fnam.size(); ++i )
            if (fnam[i]!=(fin->name)[i])
                fnam[i] = ' ';
    }
    for ( auto cp = fnam.begin(); cp != fnam.end(); ++cp )
        if (*cp==' ') fnam.erase(cp--);
    if (fnam == "") {
        static uint callNo=0;
        fnam = str( format( "merge%03u" ) % callNo++ );
    }
    //fnam = wask("Joined file name", fnam);
    //if (fnam=="")
    //    return;
    fout->name = fnam;

    // doc lines:
    fout->lDoc.push_back( "mfj # " + fnam + " is merger of:" );
    fiter.reset();
    POlo f1 = fiter();
    fiter.reset();
    while ( (fin = fiter()) ) {
        fout->lDoc.push_back("  " + fin->name);
        if ( fiter.iteration()==0 ) {
            for ( uint i=0; i<fin->lDoc.size(); ++i )
                fout->lDoc.push_back("    " + fin->lDoc[i]);
        } else {
            string line = "    ";
            for ( uint i=0; i<fin->lDoc.size(); ++i ) {
                if ( i<f1->lDoc.size() && fin->lDoc[i]==f1->lDoc[i] ) {
                    line += "=";
                } else {
                    fout->lDoc.push_back( line + fin->lDoc[i] );
                }
            }
        }
    }

    // now merge the Zentries, and add new z's:
    fiter.reset();
    while ( fin = fiter() ) {
        for ( uint j=0; j<fin->nJ(); j++ ) {
            PZentry eout( fin->new_zentry(j) );
            auto zinsert=eout->z.begin();
            for ( uint i=0; i<nR; ++i )
                if ( RParDiff[i] )
                    eout->z.insert( zinsert++, fin->RPar[i].val );
            if ( add_zk )
                eout->z.insert( eout->z.begin(), fiter.iteration() );
            fout->V.push_back( eout );
        }
    }

    if ( NOlm::overwrite ) {
        fiter.reset();
        fiter();
        uint kout = fiter.k();
        NOlm::mem_store( fout, kout );
        while ( fiter() )
            NOlm::mem_del( fiter.k() );
        CList FSelNew( kout );
        NOlm::sel_push( &FSelNew );
    } else {
        NOlm::mem_store( fout );
    }
}


//! Merge files with same number of spectra, concatenating spectra with same j.

void NManip::FilMergePointwise()
{
    NOlm::IterateD fiter;
    // Copy first input file:
    POld fin = fiter();
    POld fout( fin->new_old() );
    fout->copy_mainvec( fin );

    // Merge further input files:
    while ( fin = fiter() ) {
        if ( fin->nJ() != fout->nJ() )
            throw "different # spectra";
        if ( fin->xco != fout->xco || fin->yco != fout->yco )
            throw "different x/y coordinates";
		
        for ( uint j=0; j<fout->nJ(); ++j ) {
            PSpec sin = fin->VS(j);
            PSpec sout = fout->VS(j);
            if ( sin->z.size() != sout->z.size() )
                 throw "different # z entries";
            for ( uint i=0; i<sout->z.size(); ++i )
                if ( sin->z[i] != sout->z[i] )
                    throw "different # z values";
            if( sin->dy.size() ^ sout->dy.size() )
                throw "different on/off status of error bars";
            for ( uint i=0; i<sin->size(); ++i ) {
                if( sin->dy.size() )
                    sout->push_xyd( sin->x[i], sin->y[i], sin->dy[i] );
                else
                    sout->push_xy( sin->x[i], sin->y[i] );
            }
        }

        fout->lDoc.push_back( "appended pointwise " + fin->name );
    }

    NOlm::mem_store( fout );
}


//***************************************************************************//
//*  Interpolation                                                          *//
//***************************************************************************//

//! Locate y in x' to determine y'(x'=y).

void NManip::Interpolate()
{
    string expr = wask( "Interpolate in reference file" );
    if (expr=="")
        return;
    PNode T = user_xaxparse( expr.c_str() );

    NOlm::IterateD fiter;
    while( POld fin = fiter() ) {
        uint k = fiter.k();
        POld fout( fin->new_old() );
        fout->copy_mainvec( fin );

        uint k2;
        CContext ctx( k );
        T->tree_uival( &k2, ctx );
        POld f2 = NOlm::mem_get_D( k2 );

        fout->lDoc.push_back( "mi " + S(k2) );

        for ( uint j=0; j<fin->nJ(); j++ ) {
            PSpec sin = fin ->VS(j);
            PSpec s2  = f2  ->VS(j);
            PSpec sout= fout->VS(j);
            uint i2 = 0;
            if( sin->y[0]<s2->x[0] )
                throw "File "+S(k)+", spec "+S(j)+": y[0] not in reference";
            for ( uint i=0; i<sin->size(); i++ ) {
                double p = sin->y[i];
                if( p==s2->x[i2] ) {
                    sout->y[i] = s2->y[i2];
                    continue;
                }
                while( i2+1<s2->size() ) {
                    if( p<=s2->x[i2+1] )
                        goto found_i2;
                    ++i2;
                }
                throw "File "+S(k)+", spec "+S(j)+": y["+S(i)+
                    "] not in reference";
            found_i2:
                double step = s2->x[i2+1] - s2->x[i2];
                if( step <= 0 )
                    throw "Reference not monotonous";
                sout->y[i] = ( s2->y[i2]*(s2->x[i2+1]-p) +
                               s2->y[i2+1]*(p-s2->x[i2]) ) / step;
            }
        }
        NOlm::mem_store( fout, fiter.k() );
    }
}
